@@ -11,19 +11,16 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
-	"os/exec"
 	"path/filepath"
 	"runtime"
+	"runtime/pprof"
 	"sort"
 	"strings"
-	"time"
 
 	"golang.org/x/mod/modfile"
 	"golang.org/x/tools/go/ast/astutil"
 	"golang.org/x/tools/gopls/internal/bug"
-	"golang.org/x/tools/gopls/internal/govulncheck"
 	"golang.org/x/tools/gopls/internal/lsp/cache"
 	"golang.org/x/tools/gopls/internal/lsp/command"
 	"golang.org/x/tools/gopls/internal/lsp/debug"
@@ -31,7 +28,9 @@ import (
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/span"
+	"golang.org/x/tools/gopls/internal/telemetry"
 	"golang.org/x/tools/gopls/internal/vulncheck"
+	"golang.org/x/tools/gopls/internal/vulncheck/scan"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/gocommand"
 	"golang.org/x/tools/internal/tokeninternal"
@@ -43,7 +42,7 @@ func (s *Server) executeCommand(ctx context.Context, params *protocol.ExecuteCom
 	defer done()
 
 	var found bool
-	for _, name := range s.session.Options().SupportedCommands {
+	for _, name := range s.Options().SupportedCommands {
 		if name == params.Command {
 			found = true
 			break
@@ -65,6 +64,20 @@ type commandHandler struct {
 	params *protocol.ExecuteCommandParams
 }
 
+func (h *commandHandler) MaybePromptForTelemetry(ctx context.Context) error {
+	go h.s.maybePromptForTelemetry(ctx, true)
+	return nil
+}
+
+func (*commandHandler) AddTelemetryCounters(_ context.Context, args command.AddTelemetryCountersArgs) error {
+	if len(args.Names) != len(args.Values) {
+		return fmt.Errorf("Names and Values must have the same length")
+	}
+	// invalid counter update requests will be silently dropped. (no audience)
+	telemetry.AddForwardedCounters(args.Names, args.Values)
+	return nil
+}
+
 // commandConfig configures common command set-up and execution.
 type commandConfig struct {
 	async       bool                 // whether to run the command asynchronously. Async commands can only return errors.
@@ -96,7 +109,7 @@ func (c *commandHandler) run(ctx context.Context, cfg commandConfig, run command
 	if cfg.requireSave {
 		var unsaved []string
 		for _, overlay := range c.s.session.Overlays() {
-			if !overlay.Saved() {
+			if !overlay.SameContentsOnDisk() {
 				unsaved = append(unsaved, overlay.URI().Filename())
 			}
 		}
@@ -426,7 +439,7 @@ func dropDependency(snapshot source.Snapshot, pm *source.ParsedModule, modulePat
 		return nil, err
 	}
 	// Calculate the edits to be made due to the change.
-	diff := snapshot.View().Options().ComputeEdits(string(pm.Mapper.Content), string(newContent))
+	diff := snapshot.Options().ComputeEdits(string(pm.Mapper.Content), string(newContent))
 	return source.ToProtocolEdits(pm.Mapper, diff)
 }
 
@@ -628,12 +641,12 @@ func collectFileEdits(ctx context.Context, snapshot source.Snapshot, uri span.UR
 	// file and leave it unsaved. We would rather apply the changes directly,
 	// especially to go.sum, which should be mostly invisible to the user.
 	if !snapshot.IsOpen(uri) {
-		err := ioutil.WriteFile(uri.Filename(), newContent, 0666)
+		err := os.WriteFile(uri.Filename(), newContent, 0666)
 		return nil, err
 	}
 
 	m := protocol.NewMapper(fh.URI(), oldContent)
-	diff := snapshot.View().Options().ComputeEdits(string(oldContent), string(newContent))
+	diff := snapshot.Options().ComputeEdits(string(oldContent), string(newContent))
 	edits, err := source.ToProtocolEdits(m, diff)
 	if err != nil {
 		return nil, err
@@ -839,6 +852,49 @@ func (c *commandHandler) StartDebugging(ctx context.Context, args command.Debugg
 		return result, fmt.Errorf("starting debug server: %w", err)
 	}
 	result.URLs = []string{"http://" + listenedAddr}
+	openClientBrowser(ctx, c.s.client, result.URLs[0])
+	return result, nil
+}
+
+func (c *commandHandler) StartProfile(ctx context.Context, args command.StartProfileArgs) (result command.StartProfileResult, _ error) {
+	file, err := os.CreateTemp("", "gopls-profile-*")
+	if err != nil {
+		return result, fmt.Errorf("creating temp profile file: %v", err)
+	}
+
+	c.s.ongoingProfileMu.Lock()
+	defer c.s.ongoingProfileMu.Unlock()
+
+	if c.s.ongoingProfile != nil {
+		file.Close() // ignore error
+		return result, fmt.Errorf("profile already started (for %q)", c.s.ongoingProfile.Name())
+	}
+
+	if err := pprof.StartCPUProfile(file); err != nil {
+		file.Close() // ignore error
+		return result, fmt.Errorf("starting profile: %v", err)
+	}
+
+	c.s.ongoingProfile = file
+	return result, nil
+}
+
+func (c *commandHandler) StopProfile(ctx context.Context, args command.StopProfileArgs) (result command.StopProfileResult, _ error) {
+	c.s.ongoingProfileMu.Lock()
+	defer c.s.ongoingProfileMu.Unlock()
+
+	prof := c.s.ongoingProfile
+	c.s.ongoingProfile = nil
+
+	if prof == nil {
+		return result, fmt.Errorf("no ongoing profile")
+	}
+
+	pprof.StopCPUProfile()
+	if err := prof.Close(); err != nil {
+		return result, fmt.Errorf("closing profile file: %v", err)
+	}
+	result.File = prof.Name()
 	return result, nil
 }
 
@@ -853,10 +909,10 @@ type pkgLoadConfig struct {
 	Tests bool
 }
 
-func (c *commandHandler) FetchVulncheckResult(ctx context.Context, arg command.URIArg) (map[protocol.DocumentURI]*govulncheck.Result, error) {
-	ret := map[protocol.DocumentURI]*govulncheck.Result{}
+func (c *commandHandler) FetchVulncheckResult(ctx context.Context, arg command.URIArg) (map[protocol.DocumentURI]*vulncheck.Result, error) {
+	ret := map[protocol.DocumentURI]*vulncheck.Result{}
 	err := c.run(ctx, commandConfig{forURI: arg.URI}, func(ctx context.Context, deps commandDeps) error {
-		if deps.snapshot.View().Options().Vulncheck == source.ModeVulncheckImports {
+		if deps.snapshot.Options().Vulncheck == source.ModeVulncheckImports {
 			for _, modfile := range deps.snapshot.ModFiles() {
 				res, err := deps.snapshot.ModVuln(ctx, modfile)
 				if err != nil {
@@ -893,60 +949,22 @@ func (c *commandHandler) RunGovulncheck(ctx context.Context, args command.Vulnch
 	}, func(ctx context.Context, deps commandDeps) error {
 		tokenChan <- deps.work.Token()
 
-		view := deps.snapshot.View()
-		opts := view.Options()
-		// quickly test if gopls is compiled to support govulncheck
-		// by checking vulncheck.Main. Alternatively, we can continue and
-		// let the `gopls vulncheck` command fail. This is lighter-weight.
-		if vulncheck.Main == nil {
-			return errors.New("vulncheck feature is not available")
-		}
+		workDoneWriter := progress.NewWorkDoneWriter(ctx, deps.work)
+		dir := filepath.Dir(args.URI.SpanURI().Filename())
+		pattern := args.Pattern
 
-		cmd := exec.CommandContext(ctx, os.Args[0], "vulncheck", "-config", args.Pattern)
-		cmd.Dir = filepath.Dir(args.URI.SpanURI().Filename())
-
-		var viewEnv []string
-		if e := opts.EnvSlice(); e != nil {
-			viewEnv = append(os.Environ(), e...)
-		}
-		cmd.Env = viewEnv
-
-		// stdin: gopls vulncheck expects JSON-encoded configuration from STDIN when -config flag is set.
-		var stdin bytes.Buffer
-		cmd.Stdin = &stdin
-
-		if err := json.NewEncoder(&stdin).Encode(pkgLoadConfig{
-			BuildFlags: opts.BuildFlags,
-			// TODO(hyangah): add `tests` flag in command.VulncheckArgs
-		}); err != nil {
-			return fmt.Errorf("failed to pass package load config: %v", err)
-		}
-
-		// stderr: stream gopls vulncheck's STDERR as progress reports
-		er := progress.NewEventWriter(ctx, "vulncheck")
-		stderr := io.MultiWriter(er, progress.NewWorkDoneWriter(ctx, deps.work))
-		cmd.Stderr = stderr
-		// TODO: can we stream stdout?
-		stdout, err := cmd.Output()
+		result, err := scan.RunGovulncheck(ctx, pattern, deps.snapshot, dir, workDoneWriter)
 		if err != nil {
-			return fmt.Errorf("failed to run govulncheck: %v", err)
+			return err
 		}
 
-		var result govulncheck.Result
-		if err := json.Unmarshal(stdout, &result); err != nil {
-			// TODO: for easy debugging, log the failed stdout somewhere?
-			return fmt.Errorf("failed to parse govulncheck output: %v", err)
-		}
-		result.Mode = govulncheck.ModeGovulncheck
-		result.AsOf = time.Now()
-		deps.snapshot.View().SetVulnerabilities(args.URI.SpanURI(), &result)
-
+		deps.snapshot.View().SetVulnerabilities(args.URI.SpanURI(), result)
 		c.s.diagnoseSnapshot(deps.snapshot, nil, false, 0)
-		vulns := result.Vulns
-		affecting := make([]string, 0, len(vulns))
-		for _, v := range vulns {
-			if v.IsCalled() {
-				affecting = append(affecting, v.OSV.ID)
+
+		affecting := make(map[string]bool, len(result.Entries))
+		for _, finding := range result.Findings {
+			if len(finding.Trace) > 1 { // at least 2 frames if callstack exists (vulnerability, entry)
+				affecting[finding.OSV] = true
 			}
 		}
 		if len(affecting) == 0 {
@@ -955,10 +973,14 @@ func (c *commandHandler) RunGovulncheck(ctx context.Context, args command.Vulnch
 				Message: "No vulnerabilities found",
 			})
 		}
-		sort.Strings(affecting)
+		affectingOSVs := make([]string, 0, len(affecting))
+		for id := range affecting {
+			affectingOSVs = append(affectingOSVs, id)
+		}
+		sort.Strings(affectingOSVs)
 		return c.s.client.ShowMessage(ctx, &protocol.ShowMessageParams{
 			Type:    protocol.Warning,
-			Message: fmt.Sprintf("Found %v", strings.Join(affecting, ", ")),
+			Message: fmt.Sprintf("Found %v", strings.Join(affectingOSVs, ", ")),
 		})
 	})
 	if err != nil {
@@ -1113,7 +1135,7 @@ func (c *commandHandler) RunGoWorkCommand(ctx context.Context, args command.RunG
 			if err != nil {
 				return fmt.Errorf("reading current go.work file: %v", err)
 			}
-			if !fh.Saved() {
+			if !fh.SameContentsOnDisk() {
 				return fmt.Errorf("must save workspace file %s before running go work commands", goworkURI)
 			}
 		} else {
@@ -1144,3 +1166,69 @@ func (c *commandHandler) invokeGoWork(ctx context.Context, viewDir, gowork strin
 	}
 	return nil
 }
+
+// openClientBrowser causes the LSP client to open the specified URL
+// in an external browser.
+func openClientBrowser(ctx context.Context, cli protocol.Client, url protocol.URI) {
+	showDocumentImpl(ctx, cli, url, nil)
+}
+
+// openClientEditor causes the LSP client to open the specified document
+// and select the indicated range.
+func openClientEditor(ctx context.Context, cli protocol.Client, loc protocol.Location) {
+	showDocumentImpl(ctx, cli, protocol.URI(loc.URI), &loc.Range)
+}
+
+func showDocumentImpl(ctx context.Context, cli protocol.Client, url protocol.URI, rangeOpt *protocol.Range) {
+	// In principle we shouldn't send a showDocument request to a
+	// client that doesn't support it, as reported by
+	// ShowDocumentClientCapabilities. But even clients that do
+	// support it may defer the real work of opening the document
+	// asynchronously, to avoid deadlocks due to rentrancy.
+	//
+	// For example: client sends request to server; server sends
+	// showDocument to client; client opens editor; editor causes
+	// new RPC to be sent to server, which is still busy with
+	// previous request. (This happens in eglot.)
+	//
+	// So we can't rely on the success/failure information.
+	// That's the reason this function doesn't return an error.
+
+	// "External" means run the system-wide handler (e.g. open(1)
+	// on macOS or xdg-open(1) on Linux) for this URL, ignoring
+	// TakeFocus and Selection. Note that this may still end up
+	// opening the same editor (e.g. VSCode) for a file: URL.
+	res, err := cli.ShowDocument(ctx, &protocol.ShowDocumentParams{
+		URI:       url,
+		External:  rangeOpt == nil,
+		TakeFocus: true,
+		Selection: rangeOpt, // optional
+	})
+	if err != nil {
+		event.Error(ctx, "client.showDocument: %v", err)
+	} else if res != nil && !res.Success {
+		event.Log(ctx, fmt.Sprintf("client declined to open document %v", url))
+	}
+}
+
+func (c *commandHandler) ChangeSignature(ctx context.Context, args command.ChangeSignatureArgs) error {
+	return c.run(ctx, commandConfig{
+		forURI: args.RemoveParameter.URI,
+	}, func(ctx context.Context, deps commandDeps) error {
+		// For now, gopls only supports removing unused parameters.
+		changes, err := source.RemoveUnusedParameter(ctx, deps.fh, args.RemoveParameter.Range, deps.snapshot)
+		if err != nil {
+			return err
+		}
+		r, err := c.s.client.ApplyEdit(ctx, &protocol.ApplyWorkspaceEditParams{
+			Edit: protocol.WorkspaceEdit{
+				DocumentChanges: changes,
+			},
+		})
+		if !r.Applied {
+			return fmt.Errorf("failed to apply edits: %v", r.FailureReason)
+		}
+
+		return nil
+	})
+}
