@@ -11,6 +11,7 @@ import (
 	"fmt"
 	"os"
 	"path/filepath"
+	"sort"
 	"strings"
 	"sync"
 	"time"
@@ -141,6 +142,8 @@ func computeDiagnosticHash(diags ...*source.Diagnostic) string {
 		for _, r := range d.Related {
 			fmt.Fprintf(h, "related: %s %s %s\n", r.Location.URI.SpanURI(), r.Message, r.Location.Range)
 		}
+		fmt.Fprintf(h, "code: %s\n", d.Code)
+		fmt.Fprintf(h, "codeHref: %s\n", d.CodeHref)
 		fmt.Fprintf(h, "message: %s\n", d.Message)
 		fmt.Fprintf(h, "range: %s\n", d.Range)
 		fmt.Fprintf(h, "severity: %s\n", d.Severity)
@@ -158,7 +161,7 @@ func (s *Server) diagnoseSnapshots(snapshots map[source.Snapshot][]span.URI, onD
 		diagnosticWG.Add(1)
 		go func(snapshot source.Snapshot, uris []span.URI) {
 			defer diagnosticWG.Done()
-			s.diagnoseSnapshot(snapshot, uris, onDisk, snapshot.View().Options().DiagnosticsDelay)
+			s.diagnoseSnapshot(snapshot, uris, onDisk, snapshot.Options().DiagnosticsDelay)
 		}(snapshot, uris)
 	}
 	diagnosticWG.Wait()
@@ -185,9 +188,27 @@ func (s *Server) diagnoseSnapshot(snapshot source.Snapshot, changedURIs []span.U
 		// file modifications.
 		//
 		// The second phase runs after the delay, and does everything.
+		//
+		// We wait a brief delay before the first phase, to allow higher priority
+		// work such as autocompletion to acquire the type checking mutex (though
+		// typically both diagnosing changed files and performing autocompletion
+		// will be doing the same work: recomputing active packages).
+		const minDelay = 20 * time.Millisecond
+		select {
+		case <-time.After(minDelay):
+		case <-ctx.Done():
+			return
+		}
+
 		s.diagnoseChangedFiles(ctx, snapshot, changedURIs, onDisk)
 		s.publishDiagnostics(ctx, false, snapshot)
 
+		if delay < minDelay {
+			delay = 0
+		} else {
+			delay -= minDelay
+		}
+
 		select {
 		case <-time.After(delay):
 		case <-ctx.Done():
@@ -227,7 +248,7 @@ func (s *Server) diagnoseChangedFiles(ctx context.Context, snapshot source.Snaps
 		}
 
 		// Find all packages that include this file and diagnose them in parallel.
-		metas, err := snapshot.MetadataForFile(ctx, uri)
+		meta, err := source.NarrowestMetadataForFile(ctx, snapshot, uri)
 		if err != nil {
 			if ctx.Err() != nil {
 				return
@@ -237,10 +258,7 @@ func (s *Server) diagnoseChangedFiles(ctx context.Context, snapshot source.Snaps
 			// noisy to log (and we'll handle things later in the slow pass).
 			continue
 		}
-		source.RemoveIntermediateTestVariants(&metas)
-		for _, m := range metas {
-			toDiagnose[m.ID] = m
-		}
+		toDiagnose[meta.ID] = meta
 	}
 	s.diagnosePkgs(ctx, snapshot, toDiagnose, nil)
 }
@@ -263,7 +281,7 @@ func (s *Server) diagnose(ctx context.Context, snapshot source.Snapshot, analyze
 	// Wait for a free diagnostics slot.
 	// TODO(adonovan): opt: shouldn't it be the analysis implementation's
 	// job to de-dup and limit resource consumption? In any case this
-	// this function spends most its time waiting for awaitLoaded, at
+	// function spends most its time waiting for awaitLoaded, at
 	// least initially.
 	select {
 	case <-ctx.Done():
@@ -362,7 +380,7 @@ func (s *Server) diagnose(ctx context.Context, snapshot source.Snapshot, analyze
 	var (
 		seen       = map[span.URI]struct{}{}
 		toDiagnose = make(map[source.PackageID]*source.Metadata)
-		toAnalyze  = make(map[source.PackageID]*source.Metadata)
+		toAnalyze  = make(map[source.PackageID]unit)
 	)
 	for _, m := range workspace {
 		var hasNonIgnored, hasOpenFile bool
@@ -378,7 +396,7 @@ func (s *Server) diagnose(ctx context.Context, snapshot source.Snapshot, analyze
 		if hasNonIgnored {
 			toDiagnose[m.ID] = m
 			if analyze == analyzeEverything || analyze == analyzeOpenPackages && hasOpenFile {
-				toAnalyze[m.ID] = m
+				toAnalyze[m.ID] = unit{}
 			}
 		}
 	}
@@ -416,7 +434,7 @@ func (s *Server) diagnose(ctx context.Context, snapshot source.Snapshot, analyze
 // of concurrent dispatch: as of writing we concurrently run TidyDiagnostics
 // and diagnosePkgs, and diagnosePkgs concurrently runs PackageDiagnostics and
 // analysis.
-func (s *Server) diagnosePkgs(ctx context.Context, snapshot source.Snapshot, toDiagnose, toAnalyze map[source.PackageID]*source.Metadata) {
+func (s *Server) diagnosePkgs(ctx context.Context, snapshot source.Snapshot, toDiagnose map[source.PackageID]*source.Metadata, toAnalyze map[source.PackageID]unit) {
 	ctx, done := event.Start(ctx, "Server.diagnosePkgs", source.SnapshotLabels(snapshot)...)
 	defer done()
 
@@ -425,7 +443,6 @@ func (s *Server) diagnosePkgs(ctx context.Context, snapshot source.Snapshot, toD
 	var (
 		wg            sync.WaitGroup
 		pkgDiags      map[span.URI][]*source.Diagnostic
-		analysisMu    sync.Mutex
 		analysisDiags = make(map[span.URI][]*source.Diagnostic)
 	)
 
@@ -446,28 +463,30 @@ func (s *Server) diagnosePkgs(ctx context.Context, snapshot source.Snapshot, toD
 
 	// Get diagnostics from analysis framework.
 	// This includes type-error analyzers, which suggest fixes to compiler errors.
-	//
-	// TODO(adonovan): in many cases we will be analyze multiple open variants of
-	// an open package, which have significantly overlapping import graphs.
-	// It may make sense to change the Analyze API to accept a slice of IDs, or
-	// merge analysis with type-checking.
-	for _, m := range toAnalyze {
-		m := m
-		wg.Add(1)
-		go func() {
-			defer wg.Done()
-			diags, err := source.Analyze(ctx, snapshot, m.ID, false)
-			if err != nil {
-				event.Error(ctx, "warning: analyzing package", err, append(source.SnapshotLabels(snapshot), tag.Package.Of(string(m.ID)))...)
-				return
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		diags, err := source.Analyze(ctx, snapshot, toAnalyze, s.progress)
+		if err != nil {
+			var tagStr string // sorted comma-separated list of package IDs
+			{
+				// TODO(adonovan): replace with a generic map[S]any -> string
+				// function in the tag package, and use  maps.Keys + slices.Sort.
+				keys := make([]string, 0, len(toDiagnose))
+				for id := range toDiagnose {
+					keys = append(keys, string(id))
+				}
+				sort.Strings(keys)
+				tagStr = strings.Join(keys, ",")
 			}
-			analysisMu.Lock()
-			for uri, diags := range diags {
-				analysisDiags[uri] = append(analysisDiags[uri], diags...)
-			}
-			analysisMu.Unlock()
-		}()
-	}
+			event.Error(ctx, "warning: analyzing package", err, append(source.SnapshotLabels(snapshot), tag.Package.Of(tagStr))...)
+			return
+		}
+		for uri, diags := range diags {
+			analysisDiags[uri] = append(analysisDiags[uri], diags...)
+		}
+	}()
+
 	wg.Wait()
 
 	// TODO(rfindley): remove the guards against snapshot.IsBuiltin, after the
@@ -554,7 +573,7 @@ func (s *Server) diagnosePkgs(ctx context.Context, snapshot source.Snapshot, toD
 				fh := snapshot.FindFile(uri)
 				// Don't publish gc details for unsaved buffers, since the underlying
 				// logic operates on the file on disk.
-				if fh == nil || !fh.Saved() {
+				if fh == nil || !fh.SameContentsOnDisk() {
 					continue
 				}
 				s.storeDiagnostics(snapshot, uri, gcDetailsSource, diags, true)
