@@ -7,6 +7,7 @@ package fake
 import (
 	"bytes"
 	"context"
+	"encoding/json"
 	"errors"
 	"fmt"
 	"os"
@@ -43,7 +44,8 @@ type Editor struct {
 	config             EditorConfig                // editor configuration
 	buffers            map[string]buffer           // open buffers (relative path -> buffer content)
 	serverCapabilities protocol.ServerCapabilities // capabilities / options
-	watchPatterns      []*glob.Glob                // glob patterns to watch
+	semTokOpts         protocol.SemanticTokensOptions
+	watchPatterns      []*glob.Glob // glob patterns to watch
 
 	// Call metrics for the purpose of expectations. This is done in an ad-hoc
 	// manner for now. Perhaps in the future we should do something more
@@ -55,7 +57,7 @@ type Editor struct {
 
 // CallCounts tracks the number of protocol notifications of different types.
 type CallCounts struct {
-	DidOpen, DidChange, DidSave, DidChangeWatchedFiles, DidClose uint64
+	DidOpen, DidChange, DidSave, DidChangeWatchedFiles, DidClose, DidChangeConfiguration uint64
 }
 
 // buffer holds information about an open buffer in the editor.
@@ -109,6 +111,17 @@ type EditorConfig struct {
 
 	// Settings holds user-provided configuration for the LSP server.
 	Settings map[string]interface{}
+
+	// CapabilitiesJSON holds JSON client capabilities to overlay over the
+	// editor's default client capabilities.
+	//
+	// Specifically, this JSON string will be unmarshalled into the editor's
+	// client capabilities struct, before sending to the server.
+	CapabilitiesJSON []byte
+
+	// If non-nil, MessageResponder is used to respond to ShowMessageRequest
+	// messages.
+	MessageResponder func(params *protocol.ShowMessageRequestParams) (*protocol.MessageActionItem, error)
 }
 
 // NewEditor creates a new Editor.
@@ -223,9 +236,9 @@ func makeSettings(sandbox *Sandbox, config EditorConfig) map[string]interface{}
 		// asynchronous operations being completed (such as diagnosing a snapshot).
 		"verboseWorkDoneProgress": true,
 
-		// Set a generous completion budget, so that tests don't flake because
+		// Set an unlimited completion budget, so that tests don't flake because
 		// completions are too slow.
-		"completionBudget": "10s",
+		"completionBudget": "0s",
 	}
 
 	for k, v := range config.Settings {
@@ -249,15 +262,13 @@ func (e *Editor) initialize(ctx context.Context) error {
 	}
 	params.InitializationOptions = makeSettings(e.sandbox, config)
 	params.WorkspaceFolders = makeWorkspaceFolders(e.sandbox, config.WorkspaceFolders)
+
+	// Set various client capabilities that are sought by gopls.
 	params.Capabilities.Workspace.Configuration = true // support workspace/configuration
 	params.Capabilities.Window.WorkDoneProgress = true // support window/workDoneProgress
-
-	// TODO(rfindley): set client capabilities (note from the future: why?)
-
 	params.Capabilities.TextDocument.Completion.CompletionItem.TagSupport.ValueSet = []protocol.CompletionItemTag{protocol.ComplDeprecated}
 	params.Capabilities.TextDocument.Completion.CompletionItem.SnippetSupport = true
 	params.Capabilities.TextDocument.SemanticTokens.Requests.Full.Value = true
-	// copied from lsp/semantic.go to avoid import cycle in tests
 	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = []string{
 		"namespace", "type", "class", "enum", "interface",
 		"struct", "typeParameter", "parameter", "variable", "property", "enumMember",
@@ -268,14 +279,11 @@ func (e *Editor) initialize(ctx context.Context) error {
 		"declaration", "definition", "readonly", "static",
 		"deprecated", "abstract", "async", "modification", "documentation", "defaultLibrary",
 	}
-
 	// The LSP tests have historically enabled this flag,
 	// but really we should test both ways for older editors.
 	params.Capabilities.TextDocument.DocumentSymbol.HierarchicalDocumentSymbolSupport = true
-
 	// Glob pattern watching is enabled.
 	params.Capabilities.Workspace.DidChangeWatchedFiles.DynamicRegistration = true
-
 	// "rename" operations are used for package renaming.
 	//
 	// TODO(rfindley): add support for other resource operations (create, delete, ...)
@@ -284,6 +292,12 @@ func (e *Editor) initialize(ctx context.Context) error {
 			"rename",
 		},
 	}
+	// Apply capabilities overlay.
+	if config.CapabilitiesJSON != nil {
+		if err := json.Unmarshal(config.CapabilitiesJSON, &params.Capabilities); err != nil {
+			return fmt.Errorf("unmarshalling EditorConfig.CapabilitiesJSON: %v", err)
+		}
+	}
 
 	trace := protocol.TraceValues("messages")
 	params.Trace = &trace
@@ -293,8 +307,13 @@ func (e *Editor) initialize(ctx context.Context) error {
 		if err != nil {
 			return fmt.Errorf("initialize: %w", err)
 		}
+		semTokOpts, err := marshalUnmarshal[protocol.SemanticTokensOptions](resp.Capabilities.SemanticTokensProvider)
+		if err != nil {
+			return fmt.Errorf("unmarshalling semantic tokens options: %v", err)
+		}
 		e.mu.Lock()
 		e.serverCapabilities = resp.Capabilities
+		e.semTokOpts = semTokOpts
 		e.mu.Unlock()
 
 		if err := e.Server.Initialized(ctx, &protocol.InitializedParams{}); err != nil {
@@ -305,6 +324,29 @@ func (e *Editor) initialize(ctx context.Context) error {
 	return nil
 }
 
+// marshalUnmarshal is a helper to json Marshal and then Unmarshal as a
+// different type. Used to work around cases where our protocol types are not
+// specific.
+func marshalUnmarshal[T any](v any) (T, error) {
+	var t T
+	data, err := json.Marshal(v)
+	if err != nil {
+		return t, err
+	}
+	err = json.Unmarshal(data, &t)
+	return t, err
+}
+
+// HasCommand reports whether the connected server supports the command with the given ID.
+func (e *Editor) HasCommand(id string) bool {
+	for _, command := range e.serverCapabilities.ExecuteCommandProvider.Commands {
+		if command == id {
+			return true
+		}
+	}
+	return false
+}
+
 // makeWorkspaceFolders creates a slice of workspace folders to use for
 // this editing session, based on the editor configuration.
 func makeWorkspaceFolders(sandbox *Sandbox, paths []string) (folders []protocol.WorkspaceFolder) {
@@ -775,10 +817,7 @@ func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty
 
 // GoToDefinition jumps to the definition of the symbol at the given position
 // in an open buffer. It returns the location of the resulting jump.
-//
-// TODO(rfindley): rename to "Definition", to be consistent with LSP
-// terminology.
-func (e *Editor) GoToDefinition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
+func (e *Editor) Definition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
 	if err := e.checkBufferLocation(loc); err != nil {
 		return protocol.Location{}, err
 	}
@@ -793,9 +832,9 @@ func (e *Editor) GoToDefinition(ctx context.Context, loc protocol.Location) (pro
 	return e.extractFirstLocation(ctx, resp)
 }
 
-// GoToTypeDefinition jumps to the type definition of the symbol at the given location
-// in an open buffer.
-func (e *Editor) GoToTypeDefinition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
+// TypeDefinition jumps to the type definition of the symbol at the given
+// location in an open buffer.
+func (e *Editor) TypeDefinition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
 	if err := e.checkBufferLocation(loc); err != nil {
 		return protocol.Location{}, err
 	}
@@ -1348,6 +1387,9 @@ func (e *Editor) ChangeConfiguration(ctx context.Context, newConfig EditorConfig
 		if err := e.Server.DidChangeConfiguration(ctx, &params); err != nil {
 			return err
 		}
+		e.callsMu.Lock()
+		e.calls.DidChangeConfiguration++
+		e.callsMu.Unlock()
 	}
 	return nil
 }
@@ -1469,3 +1511,61 @@ func (e *Editor) DocumentHighlight(ctx context.Context, loc protocol.Location) (
 
 	return e.Server.DocumentHighlight(ctx, params)
 }
+
+// SemanticTokens invokes textDocument/semanticTokens/full, and interprets its
+// result.
+func (e *Editor) SemanticTokens(ctx context.Context, path string) ([]SemanticToken, error) {
+	p := &protocol.SemanticTokensParams{
+		TextDocument: protocol.TextDocumentIdentifier{
+			URI: e.sandbox.Workdir.URI(path),
+		},
+	}
+	resp, err := e.Server.SemanticTokensFull(ctx, p)
+	if err != nil {
+		return nil, err
+	}
+	content, ok := e.BufferText(path)
+	if !ok {
+		return nil, fmt.Errorf("buffer %s is not open", path)
+	}
+	return e.interpretTokens(resp.Data, content), nil
+}
+
+// A SemanticToken is an interpreted semantic token value.
+type SemanticToken struct {
+	Token     string
+	TokenType string
+	Mod       string
+}
+
+// Note: previously this function elided comment, string, and number tokens.
+// Instead, filtering of token types should be done by the caller.
+func (e *Editor) interpretTokens(x []uint32, contents string) []SemanticToken {
+	e.mu.Lock()
+	legend := e.semTokOpts.Legend
+	e.mu.Unlock()
+	lines := strings.Split(contents, "\n")
+	ans := []SemanticToken{}
+	line, col := 1, 1
+	for i := 0; i < len(x); i += 5 {
+		line += int(x[i])
+		col += int(x[i+1])
+		if x[i] != 0 { // new line
+			col = int(x[i+1]) + 1 // 1-based column numbers
+		}
+		sz := x[i+2]
+		t := legend.TokenTypes[x[i+3]]
+		l := x[i+4]
+		var mods []string
+		for i, mod := range legend.TokenModifiers {
+			if l&(1<<i) != 0 {
+				mods = append(mods, mod)
+			}
+		}
+		// Preexisting note: "col is a utf-8 offset"
+		// TODO(rfindley): is that true? Or is it UTF-16, like other columns in the LSP?
+		tok := lines[line-1][col-1 : col-1+int(sz)]
+		ans = append(ans, SemanticToken{tok, t, strings.Join(mods, " ")})
+	}
+	return ans
+}
